/*
 * Copyright 2021 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.record;

import java.math.BigDecimal;
import java.util.Date;
import java.util.Map;
import java.util.logging.Logger;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtField;
import javassist.CtNewConstructor;
import javassist.CtNewMethod;
import javassist.NotFoundException;

/**
 * Class CodeGen generates FastStructReader's and FastStructWriter's using javassist library.
 */
public class CodeGen {
	private static Logger log = Logger.getLogger(CodeGen.class.getName());
	private static String cap(String s) {
		return s.substring(0, 1).toUpperCase() + s.substring(1);
	}
	private static void badConversionForRead(FieldType ft, Class<?> clz, Class<?> struct) throws RecordException {
		throw new RecordException("Can not convert from " + ft.toString() + " to " + clz.getName() + " in " + struct.getName());
	}
	private static void badConversionForWrite(Class<?> clz, FieldType ft, Class<?> struct) throws RecordException {
		throw new RecordException("Can not convert from " + clz.getName() + " to " + ft.toString() + " in " + struct.getName());
	}
	private static <T> String makeReadMethod(Class<T> t) throws RecordException {
		int selpad = 0;
		StructInfo si = StructInfoCache.analyze(t);
		String res = "public Object read(Class t) {\r\n";
		res += "    try {\r\n";
		res += String.format("        bb.order(%s);\r\n", si.getEndianess() == Endian.LITTLE ? "java.nio.ByteOrder.LITTLE_ENDIAN" : "java.nio.ByteOrder.BIG_ENDIAN");
		res += String.format("        %s res = (%s)t.newInstance();\r\n", t.getName(), t.getName()); 
		res += "        long bitbuf = 0;\r\n";
		res += "        int nbits = 0;\r\n";
	    res += "        int nelm;\r\n";
	    res += "        int nskip;\r\n";
	    res += "        java.nio.ByteOrder ctxbo = null;\r\n";
		for(int i = 0; i < si.getFields().size(); i++) {
		    FieldInfo fi = si.getFields().get(i);
		    if(fi.getEndianess() != ExtEndian.CONTEXT) {
		    	res += "        ctxbo = bb.order();\r\n";
				res += String.format("        bb.order(%s);\r\n", fi.getEndianess() == ExtEndian.LITTLE ? "java.nio.ByteOrder.LITTLE_ENDIAN" : "java.nio.ByteOrder.BIG_ENDIAN");
		    }
		    int align = StructInfo.calculateAlign(si.getAlignment(), fi);
		    if(align > 1) {
			    res += String.format("        nskip = (%d - bb.position() %% %d) %% %d;\r\n", align, align, align);
		        res += "        if(nskip > 0) bb.get(new byte[nskip]);\r\n";
		    }
	    	if(fi.getCountprefix() <= 0) {
	    		res += String.format("        nelm = %d;\r\n", fi.getElements());
	    	} else {
	    		switch(fi.getCountprefix()) {
	    			case 1:
	    				res += "        nelm = bb.get();\r\n";
	    				break;
	    			case 2:
	    				res += "        nelm = bb.getShort();\r\n";
	    				break;
	    			case 4:
	    				res += "        nelm = bb.getInt();\r\n";
	    				break;
	    			default:
	    				throw new RecordException("Wrong length of count prefix " + fi.getField().getName() + " in " + t.getName() + ": " + fi.getCountprefix());
	    		}
		    }
	    	boolean single = fi.getCountprefix() <= 0 && fi.getElements() == 1;
	    	if(!single) {
	    		res += "        for(int ix = 0; ix < nelm; ix++) {\r\n";
	    	}
	    	switch(fi.getStructType()) {
				case INT1:
					res += "            byte vint1 = bb.get();\r\n";
					if(fi.getClassType() == byte.class ||
					   fi.getClassType() == short.class ||
					   fi.getClassType() == int.class) {
						res += String.format("            res.set%s(vint1);\r\n", cap(fi.getField().getName()));
	                } else if(fi.getClassType() == byte[].class) {
	                	res += String.format("            res.get%s()[ix] = vint1;\r\n", cap(fi.getField().getName()));
					} else {
						badConversionForRead(fi.getStructType(), fi.getClassType(), t);
					}
					break;
				case INT2:
					res += "            short vint2 = bb.getShort();\r\n";
					if(fi.getClassType() == short.class ||
					   fi.getClassType() == int.class) {
						res += String.format("            res.set%s(vint2);\r\n", cap(fi.getField().getName()));
	                } else if(fi.getClassType() == short[].class) {
	                	res += String.format("            res.get%s()[ix] = vint2;\r\n", cap(fi.getField().getName()));
					} else {
						badConversionForRead(fi.getStructType(), fi.getClassType(), t);
					}
					break;
				case INT4:
					res += "            int vint4 = bb.getInt();\r\n";
					if(fi.getClassType() == int.class) {
						res += String.format("            res.set%s(vint4);\r\n", cap(fi.getField().getName()));
	                } else if(fi.getClassType() == int[].class) {
	                	res += String.format("            res.get%s()[ix] = vint4;\r\n", cap(fi.getField().getName()));
					} else {
						badConversionForRead(fi.getStructType(), fi.getClassType(), t);
					}
					break;
				case INT8:
					res += "            long vint8 = bb.getLong();\r\n";
					if(fi.getClassType() == long.class) {
						res += String.format("            res.set%s(vint8);\r\n", cap(fi.getField().getName()));
	                } else if(fi.getClassType() == long[].class) {
	                	res += String.format("            res.get%s()[ix] = vint8;\r\n", cap(fi.getField().getName()));
					} else {
						badConversionForRead(fi.getStructType(), fi.getClassType(), t);
					}
					break;
                case UINT1:
                    res += "            short vuint1 = (short)(0xFF & bb.get());\r\n";
                    if(fi.getClassType() == short.class ||
                       fi.getClassType() == int.class) {
                    	res += String.format("            res.set%s(vuint1);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == short[].class) {
                    	res += String.format("            res.get%s()[ix] = vuint1;\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
                case UINT2:
                    res += "            int vuint2 = 0xFFFF & bb.getShort();\r\n";
                    if(fi.getClassType() == int.class) {
                    	res += String.format("            res.set%s(vuint2);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == int[].class) {
                    	res += String.format("            res.get%s()[ix] = v2int2;\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
                case UINT4:
                    res += "            long vuint4 = 0xFFFFFFFFL & bb.getInt();\r\n";
                    if(fi.getClassType() == long.class) {
                    	res += String.format("            res.set%s(vuint4);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == long[].class) {
                    	res += String.format("            res.get%s()[ix] = vuint4;\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
				case FP4:
					res += "            float vfp4 = bb.getFloat();\r\n";
					if(fi.getClassType() == float.class) {
						res += String.format("            res.set%s(vfp4);\r\n", cap(fi.getField().getName()));
	                } else if(fi.getClassType() == float[].class) {
	                	res += String.format("            res.get%s()[ix] = vfp4;\r\n", cap(fi.getField().getName()));
					} else {
						badConversionForRead(fi.getStructType(), fi.getClassType(), t);
					}
					break;
				case FP8:
					res += "            double vfp8 = bb.getDouble();\r\n";
					if(fi.getClassType() == double.class) {
						res += String.format("            res.set%s(vfp8);\r\n", cap(fi.getField().getName()));
	                } else if(fi.getClassType() == double[].class) {
	                	res += String.format("            res.get%s()[ix] = vfp8;\r\n", cap(fi.getField().getName()));
					} else {
						badConversionForRead(fi.getStructType(), fi.getClassType(), t);
					}
					break;
                case INTX:
                    if(fi.getLength() <= 0 || fi.getLength() >= 8) {
                        throw new RecordException("Wrong length of general integer " + fi.getField().getName() + " in " + t.getName() + ": " + fi.getLength());    
                    }
                    res += String.format("            int intxlen = %d;\r\n", fi.getLength());
                    res += "            long vintx = 0;\r\n";
                    if(si.getEndianess() == Endian.BIG) {
	                    res += "           for(int j = 0; j < intxlen; j++) {\r\n";
	                    res += "               vintx = vintx | ((bb.get() & 0x00FFL) << ((intxlen - 1 - j) * 8));\r\n";
	                    res += "            }\r\n";
                    } else if(si.getEndianess() == Endian.LITTLE) {
	                    res += "            for(int j = 0; j < intxlen; j++) {\r\n";
	                    res += "                vintx = vintx | ((bb.get() & 0x00FFL) << (j * 8));\r\n";
	                    res += "            }\r\n";
                    }
                    if(fi.getClassType() == long.class) {
                    	res += String.format("            res.set%s(vintx);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == long[].class) {
                    	res += String.format("            res.get%s()[ix] = vintx\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
				case FIXSTR:
					res += String.format("            int fixstrlen = %d;\r\n", fi.getLength());
					res += "            byte[] fixstrba = new byte[fixstrlen];\r\n";
                    res += "            bb.get(fixstrba);\r\n";
                    res += "                    int trimlen = fixstrlen;\r\n";
                    if(fi.getPad()) {
                        res += String.format("                    while(trimlen > 0 && fixstrba[trimlen - 1] == %d) trimlen--;\r\n", (int)fi.getPadchar());
                    }
                    res += String.format("            String vfixstr = new String(fixstrba, 0, trimlen, \"%s\");\r\n", fi.getEncoding());
					if(fi.getClassType() == String.class) {
						res += String.format("            res.set%s(vfixstr);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == String[].class) {
	                	res += String.format("            res.get%s()[ix] = vfixstr;\r\n", cap(fi.getField().getName()));
					} else {
						badConversionForRead(fi.getStructType(), fi.getClassType(), t);
					}
					
					break;
                case FIXSTRNULTERM:
                    res += String.format("            int fixstrnultermlen = %d;\r\n", fi.getLength());
                    res += "            byte[] fixstrnultermba = new byte[fixstrnultermlen];\r\n";
                    res += "            bb.get(fixstrnultermba);\r\n";
                    res += "            int actlen = 0;\r\n";
                    res += String.format("            while(actlen < %d && fixstrnultermba[actlen] != 0) {\r\n", fi.getLength());
                    res += "                actlen++;\r\n";
                    res += "            }\r\n";
                    res += String.format("            String vfixstrnulterm = new String(fixstrnultermba, 0, actlen, \"%s\");\r\n", fi.getEncoding());
                    if(fi.getClassType() == String.class) {
                    	res += String.format("            res.set%s(vfixstrnulterm);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == String[].class) {
                    	res += String.format("            res.get%s()[ix] = vfixstrnulterm;\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
                case VARSTR:
                    switch(fi.getPrefixlength()) {
                        case 0:
                        case 2:
                            res += "            int varstrlen = bb.getShort() & 0x0000FFFF;\r\n";
                            break;
                        case 1:
                            res += "            int varstrlen = bb.get() & 0x000000FF;\r\n";
                            break;
                        case 4:
                            res += "            int varstrlen = bb.getInt();\r\n";
                            break;
                        default:
                        	throw new RecordException("Wrong size prefix length of string " + fi.getField().getName() + " in " + t.getName() + ": " + fi.getPrefixlength());
                    }
                    res += "            byte[] varstrba = new byte[varstrlen];\r\n";
                    res += "            bb.get(varstrba);\r\n";
                    res += String.format("            String vvarstr = new String(varstrba, \"%s\");\r\n", fi.getEncoding());
                    if(fi.getClassType() == String.class) {
                    	res += String.format("            res.set%s(vvarstr);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == String[].class) {
                    	res += String.format("            res.get%s()[ix] = vvarstr;\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
                case VARFIXSTR:
                    switch(fi.getPrefixlength()) {
	                    case 0:
	                    case 2:
	                        res += "            int varfixstrlen = bb.getShort() & 0x0000FFFF;\r\n";
	                        break;
	                    case 1:
	                        res += "            int varfixstrlen = bb.get() & 0x000000FF;\r\n";
	                        break;
	                    case 4:
	                        res += "            int varfixstrlen = bb.getInt();\r\n";
	                        break;
	                    default:
	                    	throw new RecordException("Wrong size prefix length of string " + fi.getField().getName() + " in " + t.getName() + ": " + fi.getPrefixlength());
	                }
	                res += "            byte[] varfixstrba = new byte[varfixstrlen];\r\n";
	                res += "            bb.get(varfixstrba);\r\n";
	                res += String.format("            String vvarfixstr = new String(varfixstrba, \"%s\");\r\n", fi.getEncoding());
	                if(fi.getClassType() == String.class) {
	                	res += String.format("            res.set%s(vvarfixstr);\r\n", cap(fi.getField().getName()));
	                } else if(fi.getClassType() == String[].class) {
	                	res += String.format("            res.get%s()[ix] = vvarfixstr;\r\n", cap(fi.getField().getName()));
	                } else {
	                    badConversionForRead(fi.getStructType(), fi.getClassType(), t);
	                }
                    res += String.format("            byte[] zero = new byte[%d - varfixstrlen];\r\n", fi.getLength());
                    res += "            bb.get(zero);\r\n ";
                    break;
                case REMSTR:
                    res += "            int remstrlen = bb.remaining();\r\n";
                    res += "            byte[] remstrba = new byte[remstrlen];\r\n"; 
                    res += "            bb.get(remstrba);\r\n";
                    res += String.format("            String vremstr = new String(remstrba, \"%s\");\r\n", fi.getEncoding());
                    if(fi.getClassType() == String.class) {
                    	res += String.format("            res.set%s(vremstr);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == String[].class) {
                    	res += String.format("            res.get%s()[ix] = vremstr;\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
                case BOOLEAN:
                    res += String.format("            byte[] booleanba = new byte[%d];\r\n", fi.getLength());
                    res += "            bb.get(booleanba);\r\n";
                    res += "            boolean vboolean = booleanba[0] != 0;\r\n";
                    if(fi.getClassType() == boolean.class) {
                    	res += String.format("            res.set%s(vboolean);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == boolean[].class) {
                    	res += String.format("            res.get%s()[ix] = vboolean;\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
				case BIT:
                    res += String.format("            while(nbits < %d) {\r\n", fi.getLength()); 
                    res += "                bitbuf = (bitbuf << 8) | (0xFF & bb.get());\r\n";
                    res += "                nbits += 8;\r\n";
                    res += "            }\r\n";
                    res += String.format("            int vbit = (int)(bitbuf >> (nbits - %d));\r\n", fi.getLength());
                    res += String.format("            bitbuf ^= (vbit << (nbits - %d));\r\n", fi.getLength());
                    res += String.format("            nbits -= %d;\r\n", fi.getLength());
					if(fi.getClassType() == int.class) {
                    	res += String.format("            res.set%s(vbit);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == int[].class) {
                    	res += String.format("            res.get%s()[ix] = vbit;\r\n", cap(fi.getField().getName()));
					} else {
						badConversionForRead(fi.getStructType(), fi.getClassType(), t);
					}
					break;
				case JAVATIME:
                    res += "            java.util.Date vjavatime = TimeUtil.fromJavaTime(bb.getLong());\r\n";
					if(fi.getClassType() == Date.class) {
						res += String.format("            res.set%s(vjavatime);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == Date[].class) {
                    	res += String.format("            res.get%s()[ix] = vjavatime;\r\n", cap(fi.getField().getName()));
					} else {
						badConversionForRead(fi.getStructType(), fi.getClassType(), t);
					}
					break;
				case UNIXTIME:
                    res += "            java.util.Date vunixtime = TimeUtil.fromUnixTime(bb.getInt());\r\n";
					if(fi.getClassType() == Date.class) {
						res += String.format("            res.set%s(vunixtime);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == Date[].class) {
                    	res += String.format("            res.get%s()[ix] = vunixtime;\r\n", cap(fi.getField().getName()));
					} else {
						badConversionForRead(fi.getStructType(), fi.getClassType(), t);
					}
					break;
				case UNIXTIME64:
                    res += "            java.util.Date vunixtime64 = TimeUtil.fromUnixTime64(bb.getLong());\r\n";
					if(fi.getClassType() == Date.class) {
						res += String.format("            res.set%s(vunixtime64);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == Date[].class) {
                    	res += String.format("            res.get%s()[ix] = vunixtime64;\r\n", cap(fi.getField().getName()));
					} else {
						badConversionForRead(fi.getStructType(), fi.getClassType(), t);
					}
					break;
				case VMSTIME:
                    res += "            java.util.Date vvmstime = TimeUtil.fromVMSTime(bb.getLong());\r\n";
					if(fi.getClassType() == Date.class) {
						res += String.format("            res.set%s(vvmstime);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == Date[].class) {
                    	res += String.format("            res.get%s()[ix] = vvmstime;\r\n", cap(fi.getField().getName()));
					} else {
						badConversionForRead(fi.getStructType(), fi.getClassType(), t);
					}
					break;
                case PACKEDBCD:
                    res += String.format("            int packedbcdlen = %d;\r\n", fi.getLength());
                    res += "            byte[] packedbcdba = new byte[packedbcdlen];\r\n";
                    res += "            bb.get(packedbcdba);\r\n";
                    res += String.format("            java.math.BigDecimal vpackedbcd = BCDUtil.decodePackedBCD(packedbcdba, %d);\r\n", fi.getDecimals());
                    if(fi.getClassType() == BigDecimal.class) {
                    	res += String.format("            res.set%s(vpackedbcd);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == BigDecimal[].class) {
                    	res += String.format("            res.get%s()[ix] = vpackedbcd;\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
                case ZONEDBCD:
                    res += String.format("            int zonedbcdlen = %d;\r\n", fi.getLength());
                    res += "            byte[] zonedbcdba = new byte[zonedbcdlen];\r\n";
                    res += "            bb.get(zonedbcdba);\r\n";
                    res += String.format("            java.math.BigDecimal vzonedbcd = BCDUtil.decodeZonedBCD(zonedbcdba, (byte)%d, %d);\r\n", fi.getZone(), fi.getDecimals());
                    if(fi.getClassType() == BigDecimal.class) {
                    	res += String.format("            res.set%s(vzonedbcd);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == BigDecimal[].class) {
                    	res += String.format("            res.get%s()[ix] = vzonedbcd;\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
                case VAXFP4:
                    res += "            float vvaxfp4 = Float.intBitsToFloat(VAXFloatUtil.f2s(bb.getInt()));\r\n";
                    if(fi.getClassType() == float.class) {
                    	res += String.format("            res.set%s(vvaxfp4);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == float[].class) {
                    	res += String.format("            res.get%s()[ix] = vvaxfp4;\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
                case VAXFP8:
                    res += "            double vvaxfp8 = Double.longBitsToDouble(VAXFloatUtil.g2t(bb.getLong()));\r\n";
                    if(fi.getClassType() == double.class) {
                    	res += String.format("            res.set%s(vvaxfp8);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == double[].class) {
                    	res += String.format("            res.get%s()[ix] = vvaxfp8;\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
                case VAXFP8OLD:
                    res += "            double vvaxfp8old = Double.longBitsToDouble(VAXFloatUtil.d2t(bb.getLong()));\r\n";
                    if(fi.getClassType() == double.class) {
                    	res += String.format("            res.set%s(vvaxfp8old);\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == double[].class) {
                    	res += String.format("            res.get%s()[ix] = vvaxfp8old;\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForRead(fi.getStructType(), fi.getClassType(), t);
                    }
                    break;
				case STRUCT:
					throw new RecordException("Compilation not supported for fields of struct type: " +  t.getName() + "." + fi.getField().getName());
		    }
			if(fi.getSelects() != null) {
				boolean ok = false;
				int selval = Integer.MIN_VALUE;
				for(Map.Entry<Integer,SubClassAndPad> e : fi.getSelects().entrySet()) {
					if(e.getValue().getSubClass().equals(t)) {
						ok = true;
						selval = e.getKey();
						selpad = e.getValue().getPad();
					}
				}
				if(ok) {
					res += String.format("            if(res.get%s() != %d) throw new RecordException(\"Invalid subclass\");\r\n", cap(fi.getField().getName()), selval);
				} else {
					throw new RecordException("Compilation not supported for classes with subclasses: " +  t.getName() + "." + fi.getField().getName());
				}
			}
			if(!single) {
				res += "        }\r\n";
			}
		    if(fi.getEndianess() != ExtEndian.CONTEXT) {
		    	res += "        bb.order(ctxbo);\r\n";
		    }
		}
		if(selpad > 0) {
			res += String.format("        bb.get(new byte[%d]);\r\n", selpad);
		}
        if(si.getEndpad()) {
		    int align = StructInfo.calculateEndAlign(si.getAlignment(), si.getFields());
		    if(align > 1) {
			    res += String.format("        nskip = (%d - bb.position() %% %d) %% %d;\r\n", align, align, align);
		        res += "        if(nskip > 0) bb.get(new byte[nskip]);\r\n";
		    }
        }
		res += "        return res;\r\n";
	    res += "    } catch(java.nio.BufferUnderflowException e) {\r\n";
        res += "        throw new RecordException(\"Not enough bytes in input\", e);\r\n";
	    res += "    } catch (java.io.UnsupportedEncodingException e) {\r\n";
		res += "        throw new RecordException(\"Unsupported encoding for string field\", e);\r\n";
		res += "    }\r\n";
		res += "}\r\n";
		log.fine("Generate:\r\n" + res);
		return res;
	}
	/**
	 * Generate FastStructReader.
	 * @param <T> struct type
	 * @param t class of what to read
	 * @return FastStructReader implementation
	 * @throws RecordException if impossible to convert between types in class and struct
	 */
	@SuppressWarnings("unchecked")
	public static <T> FastStructReader<T> genReader(Class<T> t) throws RecordException {
        try {
        	String clznam = "dk.vajhoej.record.gen.StructReaderOf" + t.getSimpleName();
    		ClassPool cp = ClassPool.getDefault();
    		cp.importPackage("dk.vajhoej.record");
    		CtClass cc = cp.getOrNull(clznam);
    		if(cc == null) {
				cc = cp.makeClass(clznam);
				cc.addInterface(cp.get("dk.vajhoej.record.FastStructReader"));
				cc.addField(new CtField(cp.get("java.nio.ByteBuffer"), "bb", cc));
				cc.addMethod(CtNewMethod.make("public void init(byte[] ba) { bb = java.nio.ByteBuffer.wrap(ba); }", cc));
				cc.addMethod(CtNewMethod.make("public void init(java.nio.ByteBuffer bb) { this.bb = bb; }", cc));
				cc.addMethod(CtNewMethod.make( makeReadMethod(t), cc));
				cc.addMethod(CtNewMethod.make("public boolean more() { return bb.remaining() > 0; }", cc));
				cc.addMethod(CtNewMethod.make("public java.nio.ByteBuffer getByteBuffer() { return bb; }", cc));
				return (FastStructReader<T>) cc.toClass().newInstance();
    		} else {
    			return (FastStructReader<T>)Class.forName(clznam).newInstance();
    		}
		} catch (NotFoundException e) {
			throw new RecordException("Error generating code for " + t.getName() + ": " + e.getMessage());
		} catch (InstantiationException e) {
			throw new RecordException("Error generating code for " + t.getName() + ": " + e.getMessage());
		} catch (IllegalAccessException e) {
			throw new RecordException("Error generating code for " + t.getName() + ": " + e.getMessage());
		} catch (CannotCompileException e) {
			throw new RecordException("Error generating code for " + t.getName() + ": " + e.getMessage());
		} catch (ClassNotFoundException e) {
			throw new RecordException("Error generating code for " + t.getName() + ": " + e.getMessage());
		}
	}
	private static <T> String makeWriteMethod(Class<T> t) throws RecordException {
		int selpad = 0;
		String res = "public void write(Object o) {\r\n";
		res += "    try {\r\n";
		res += String.format("        %s o2 = (%s)o;\r\n", t.getName(), t.getName());
		res += "        long bitbuf = 0;\r\n";
		res += "        int nbits = 0;\r\n";
		res += "        int npad;\r\n";
	    res += "        java.nio.ByteOrder ctxbo = null;\r\n";
		StructInfo si = StructInfoCache.analyze(t);
        res += String.format("        bb.order(%s);\r\n", si.getEndianess() == Endian.LITTLE ? "java.nio.ByteOrder.LITTLE_ENDIAN" : "java.nio.ByteOrder.BIG_ENDIAN");
		for(int i = 0; i < si.getFields().size(); i++) {
		    FieldInfo fi = si.getFields().get(i);
		    if(fi.getEndianess() != ExtEndian.CONTEXT) {
		    	res += "        ctxbo = bb.order();\r\n";
				res += String.format("        bb.order(%s);\r\n", fi.getEndianess() == ExtEndian.LITTLE ? "java.nio.ByteOrder.LITTLE_ENDIAN" : "java.nio.ByteOrder.BIG_ENDIAN");
		    }
		    int align = StructInfo.calculateAlign(si.getAlignment(), fi);
		    if(align > 1) {
			    res += String.format("        npad = (%d - bb.position() %% %d) %% %d;\r\n", align, align, align);
		        res += "        if(npad > 0) bb.put(new byte[npad]);\r\n";
		    }
	    	if(fi.getCountprefix() <= 0) {
	    		res += String.format("        int nelm = %d;\r\n", fi.getElements());
	    	} else {
	    		res += String.format("        int nelm = o2.get%s().length;\r\n", cap(fi.getField().getName()));
	    		switch(fi.getCountprefix()) {
	    			case 1:
	    				res += "        bb.put((byte)nelm);\r\n";
	    				break;
	    			case 2:
	    				res += "        bb.putShort((short)nelm);\r\n";
	    				break;
	    			case 4:
	    				res += "        bb.putInt(nelm);\r\n";
	    				break;
	    			default:
	    				throw new RecordException("Wrong length of count prefix " + fi.getField().getName() + " in " + t.getName() + ": " + fi.getCountprefix());
	    		}
	    	}
	    	boolean single = fi.getCountprefix() <= 0 && fi.getElements() == 1;
	    	if(!single) {
	    		res += "        for(int ix = 0; ix < nelm; ix++) {\r\n";
	    	}
            switch(fi.getStructType()) {
	            case INT1:
	                if(fi.getClassType() == byte.class ||
	                   fi.getClassType() == short.class ||
	                   fi.getClassType() == int.class) {
	                    res += String.format("            byte vint1 = (byte)o2.get%s();\r\n", cap(fi.getField().getName()));
	                } else if(fi.getClassType() == byte[].class) {
	                    res += String.format("            byte vint1 = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
	                } else {
	                    badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
	                }
	                res += "            bb.put(vint1);\r\n";
	                break;
	            case INT2:
	                if(fi.getClassType() == short.class ||
	                   fi.getClassType() == int.class) {
	                    res += String.format("            short vint2 = (short)o2.get%s();\r\n", cap(fi.getField().getName()));
	                } else if(fi.getClassType() == int[].class) {
	                    res += String.format("            short vint2 = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
	                } else {
	                    badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
	                }
	                res += "            bb.putShort(vint2);\r\n";
	                break;
	            case INT4:
	                if(fi.getClassType() == int.class) {
	                    res += String.format("            int vint4 = o2.get%s();\r\n", cap(fi.getField().getName()));
	                } else if(fi.getClassType() == int[].class) {
	                    res += String.format("            int vint4 = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
	                } else {
	                    badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
	                }
	                res += "            bb.putInt(vint4);\r\n";
	                break;
	            case INT8:
	                if(fi.getClassType() == long.class) {
	                    res += String.format("            long vint8 = o2.get%s();\r\n", cap(fi.getField().getName()));
	                } else if(fi.getClassType() == long[].class) {
	                    res += String.format("            long vint8 = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
	                } else {
	                    badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
	                }
	                res += "            bb.putLong(vint8);\r\n";
	                break;
                case UINT1:
                    if(fi.getClassType() == short.class ||
                       fi.getClassType() == int.class) {
                    	res += String.format("            short vuint1 = (short)o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == short[].class) {
                    	res += String.format("            short vuint1 = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += "            bb.put((byte)vuint1);\r\n";
                    break;
                case UINT2:
                    if(fi.getClassType() == int.class) {
                    	res += String.format("            int vuint2 = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == int[].class) {
                    	res += String.format("            int vuint2 = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += "            bb.putShort((short)vuint2);\r\n";
                    break;
                case UINT4:
                    if(fi.getClassType() == long.class) {
                    	res += String.format("            long vuint4 = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == long[].class) {
                    	res += String.format("            long vuint4 = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += "            bb.putInt((int)vuint4);\r\n";
                    break;
                case FP4:
                    if(fi.getClassType() == float.class) {
                    	res += String.format("            float vfp4 = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == float[].class) {
                    	res += String.format("            float vfp4 = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += "            bb.putFloat(vfp4);\r\n";
                    break;
                case FP8:
                    if(fi.getClassType() == double.class) {
                    	res += String.format("            double vfp8 = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == double[].class) {
                    	res += String.format("            double vfp8 = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += "            bb.putDouble(vfp8);\r\n";
                    break;
                case INTX:
                    res += String.format("            int intxlen = %d;\r\n", fi.getLength());
                    if(fi.getClassType() == long.class) {
                    	res += String.format("            long vintx = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == long[].class) {
                    	res += String.format("            long vintx = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    if(fi.getLength() > 0 && fi.getLength() < 8) {
                        if(si.getEndianess() == Endian.BIG) {
                            res += "            for(int j = 0; j < intxlen; j++) {\r\n";
                            res += "                bb.put(((byte)((vintx >> ((intxlen - 1 - j) * 8)) & 0x00FF)));\r\n";
                            res += "            }\r\n";
                        } else if(si.getEndianess() == Endian.LITTLE) {
                            res += "            for(int j = 0; j < intxlen; j++) {\r\n";
                            res += "                bb.put(((byte)((vintx >> (j * 8)) & 0x00FF)));\r\n";
                            res += "            }\r\n";
                        }
                    } else {
                        throw new RecordException("Wrong length of general integer " + fi.getField().getName() + " in " + t.getName());
                    }
                    break;
                case FIXSTR:
                    res += String.format("            int fixstrlen = %d;\r\n", fi.getLength());
                    if(fi.getClassType() == String.class) {
                    	res += String.format("            String vfixstr = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == String[].class) {
                    	res += String.format("            String vfixstr = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += String.format("            byte[] fixstrba = vfixstr.getBytes(\"%s\");\r\n", fi.getEncoding());
                    res += "            if(fixstrba.length == fixstrlen) {\r\n";
                    res += "                bb.put(fixstrba);\r\n";
                    if(fi.getPad()) {
                        res += "            } else if(fixstrba.length < fixstrlen) {\r\n";
                        res += "                bb.put(fixstrba);\r\n";
                        res += "                byte[] pad = new byte[fixstrlen - fixstrba.length];\r\n";
                        res += String.format("                for(int j = 0; j < pad.length; j++) pad[j] = (byte)%d;\r\n", (int)fi.getPadchar());
                        res += "                bb.put(pad);\r\n";
                    }
                    res += "            } else {\r\n";
                    res += String.format("                throw new RecordException(\"Wrong length of string %s in %s\");\r\n", fi.getField().getName(), t.getName());
                    res += "            }\r\n";
                    break;
                case FIXSTRNULTERM:
                    res += String.format("            int fixstrnultermlen = %d;\r\n", fi.getLength());
                    if(fi.getClassType() == String.class) {
                    	res += String.format("            String vfixstrnulterm = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == String[].class) {
                    	res += String.format("            String vfixstrnulterm = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += String.format("            byte[] fixstrnultermba = vfixstrnulterm.getBytes(\"%s\");\r\n", fi.getEncoding());
                    res += "            if(fixstrnultermba.length <= fixstrnultermlen) {\r\n";
                    res += "                bb.put(fixstrnultermba);\r\n";
                    res += "                byte[] zero = new byte[fixstrnultermlen - fixstrnultermba.length];\r\n";
                    res += "                bb.put(zero);\r\n";
                    res += "            } else {\r\n";
                    res += String.format("                throw new RecordException(\"Wrong length of string %s in %s\");\r\n", fi.getField().getName(), t.getName());
                    res += "            }\r\n";
                    break;
                case VARSTR:
                    if(fi.getClassType() == String.class) {
                        res += String.format("            String vvarstr = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == String[].class) {
                        res += String.format("            String vvarstr = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += String.format("            byte[] varstrba = vvarstr.getBytes(\"%s\");\r\n", fi.getEncoding());
                    if((fi.getPrefixlength() == 0 || fi.getPrefixlength() == 2)) {
                    	res += "            if(varstrba.length < 32768) {\r\n";
                    	res += "                bb.putShort((short)varstrba.length);\r\n";
                    	res += "                bb.put(varstrba);\r\n";
                        res += "            } else {\r\n";
                        res += String.format("                throw new RecordException(\"Wrong length of string %s in %s\");\r\n", fi.getField().getName(), t.getName());
                        res += "            }\r\n";
                    } else if(fi.getPrefixlength() == 1) {
                    	res += "            if(varstrba.length < 128) {\r\n";
                    	res += "                bb.put((byte)varstrba.length);\r\n";
                    	res += "                bb.put(varstrba);\r\n";
                        res += "            } else {\r\n";
                        res += String.format("                throw new RecordException(\"Wrong length of string %s in %s\");\r\n", fi.getField().getName(), t.getName());
                        res += "            }\r\n";
                    } else if(fi.getPrefixlength() == 4) {
                    	res += "            bb.putInt(varstrba.length);\r\n";
                    	res += "            bb.put(varstrba);\r\n";
                    } else {
                        throw new RecordException("Invalid length prefix of string " + fi.getField().getName() + " in " + t.getName());
                    }
                    break;
                case VARFIXSTR:
                    if(fi.getClassType() == String.class) {
                        res += String.format("            String vvarfixstr = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == String[].class) {
                        res += String.format("            String vvarfixstr = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += String.format("            byte[] varfixstrba = vvarfixstr.getBytes(\"%s\");\r\n", fi.getEncoding());
                    if((fi.getPrefixlength() == 0 || fi.getPrefixlength() == 2)) {
                    	res += "            if(varfixstrba.length < 32768) {\r\n";
                    	res += "                bb.putShort((short)varfixstrba.length);\r\n";
                    	res += "                bb.put(varfixstrba);\r\n";
                        res += "            } else {\r\n";
                        res += String.format("                throw new RecordException(\"Wrong length of string %s in %s\");\r\n", fi.getField().getName(), t.getName());
                        res += "            }\r\n";
                    } else if(fi.getPrefixlength() == 1) {
                    	res += "            if(varfixstrba.length < 128) {\r\n";
                    	res += "                bb.put((byte)varfixstrba.length);\r\n";
                    	res += "                bb.put(varfixstrba);\r\n";
                        res += "            } else {\r\n";
                        res += String.format("                throw new RecordException(\"Wrong length of string %s in %s\");\r\n", fi.getField().getName(), t.getName());
                        res += "            }\r\n";
                    } else if(fi.getPrefixlength() == 4) {
                    	res += "            bb.putInt(varfixstrba.length);\r\n";
                    	res += "            bb.put(varfixstrba);\r\n";
                    } else {
                        throw new RecordException("Invalid length prefix of string " + fi.getField().getName() + " in " + t.getName());
                    }
                    res += String.format("            byte[] zero = new byte[%d - varfixstrba.length];\r\n", fi.getLength());
                    res += "            bb.put(zero);\r\n";
                    break;
                case REMSTR:
                    if(fi.getClassType() == String.class) {
                        res += String.format("            String vremstr = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == String[].class) {
                        res += String.format("            String vremstr = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += String.format("            byte[] remstrba = vremstr.getBytes(\"%s\");\r\n", fi.getEncoding());
                    res += "            bb.put(remstrba);\r\n";
                    break;
                case BOOLEAN:
                    if(fi.getClassType() == boolean.class) {
                        res += String.format("            boolean vboolean = o2.is%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == boolean[].class) {
                        res += String.format("            boolean vboolean = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += String.format("            byte[] booleanba = new byte[%d];\r\n", fi.getLength());
                    res += "            booleanba[0] =  vboolean ? (byte)1 : (byte)0;\r\n";
                    res += "            bb.put(booleanba);\r\n";
                    break;
                case BIT:
                    if(fi.getClassType() == int.class) {
                        res += String.format("            int vbit = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == int[].class) {
                        res += String.format("            int vbit = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += String.format("            bitbuf = bitbuf << %d | vbit;\r\n", fi.getLength());
                    res += String.format("            nbits += %d;\r\n", fi.getLength());
                    res += "            while(nbits >= 8) {\r\n";
                    res += "                byte tmp = (byte)(bitbuf >> (nbits - 8));\r\n";
                    res += "                bb.put(tmp);\r\n";
                    res += "                bitbuf ^= (tmp << (nbits - 8));\r\n";
                    res += "                nbits -= 8;\r\n";
                    res += "            }\r\n";
                    break;
                case JAVATIME:
                    if(fi.getClassType() == Date.class) {
                        res += String.format("            java.util.Date vjavatime = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == Date[].class) {
                        res += String.format("            java.util.Date vjavatime = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += "            bb.putLong(TimeUtil.toJavaTime(vjavatime));\r\n";
                    break;
                case UNIXTIME:
                    if(fi.getClassType() == Date.class) {
                        res += String.format("            java.util.Date vunixtime = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == Date[].class) {
                        res += String.format("            java.util.Date vunixtime = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += "            bb.putInt(TimeUtil.toUnixTime(vunixtime));\r\n";
                    break;
                case UNIXTIME64:
                    if(fi.getClassType() == Date.class) {
                        res += String.format("            java.util.Date vunixtime64 = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == Date[].class) {
                        res += String.format("            java.util.Date vunixtime64 = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += "            bb.putLong(TimeUtil.toUnixTime64(vunixtime64));\r\n";
                    break;
                case VMSTIME:
                    if(fi.getClassType() == Date.class) {
                        res += String.format("            java.util.Date vvmstime = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == Date[].class) {
                        res += String.format("            java.util.Date vvmstime = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += "            bb.putLong(TimeUtil.toVMSTime(vvmstime));\r\n";
                    break;
                case PACKEDBCD:
                    res += String.format("            int packedbcdlen = %d;\r\n", fi.getLength());
                    if(fi.getClassType() == BigDecimal.class) {
                        res += String.format("            java.math.BigDecimal vpackedbcd = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == BigDecimal[].class) {
                        res += String.format("            java.math.BigDecimal vpackedbcd = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += String.format("            bb.put(BCDUtil.encodePackedBCD(vpackedbcd, %d, packedbcdlen));\r\n", fi.getDecimals());
                    break;
                case ZONEDBCD:
                    res += String.format("            int zonedbcdlen = %d;\r\n", fi.getLength());
                    if(fi.getClassType() == BigDecimal.class) {
                        res += String.format("            java.math.BigDecimal vzonedbcd = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == BigDecimal[].class) {
                        res += String.format("            java.math.BigDecimal vzonedbcd = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += String.format("            bb.put(BCDUtil.encodeZonedBCD(vzonedbcd, (byte)%d, %d, zonedbcdlen));\r\n", fi.getZone(), fi.getDecimals());
                    break;
                case VAXFP4:
                    if(fi.getClassType() == float.class) {
                    	res += String.format("            float vvaxfp4 = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == float[].class) {
                    	res += String.format("            float vvaxfp4 = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += "            bb.putInt(VAXFloatUtil.s2f(Float.floatToRawIntBits(vvaxfp4)));\r\n";
                    break;
                case VAXFP8:
                    if(fi.getClassType() == double.class) {
                    	res += String.format("            double vvaxfp8 = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == double[].class) {
                    	res += String.format("            double vvaxfp8 = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += "            bb.putLong(VAXFloatUtil.t2g(Double.doubleToRawLongBits(vvaxfp8)));\r\n";
                    break;
                case VAXFP8OLD:
                    if(fi.getClassType() == double.class) {
                    	res += String.format("            double vvaxfp8old = o2.get%s();\r\n", cap(fi.getField().getName()));
                    } else if(fi.getClassType() == double[].class) {
                    	res += String.format("            double vvaxfp8old = o2.get%s()[ix];\r\n", cap(fi.getField().getName()));
                    } else {
                        badConversionForWrite(fi.getClassType(), fi.getStructType(), t);
                    }
                    res += "            bb.putLong(VAXFloatUtil.t2d(Double.doubleToRawLongBits(vvaxfp8old)));\r\n";
                    break;
				case STRUCT:
					throw new RecordException("Compilation not supported for fields of struct type: " +  t.getName() + "." + fi.getField().getName());
		    }
			if(fi.getSelects() != null) {
				boolean ok = false;
				int selval = Integer.MIN_VALUE;
				for(Map.Entry<Integer,SubClassAndPad> e : fi.getSelects().entrySet()) {
					if(e.getValue().getSubClass().equals(t)) {
						ok = true;
						selval = e.getKey();
						selpad = e.getValue().getPad();
					}
				}
				if(ok) {
					res += String.format("            if(o2.get%s() != %d) throw new RecordException(\"Invalid subclass\");\r\n", cap(fi.getField().getName()), selval);
				} else {
					throw new RecordException("Compilation not supported for classes with subclasses: " +  t.getName() + "." + fi.getField().getName());
				}
			}
	        if(!single) {
	        	res += "        }\r\n";
	        }
		    if(fi.getEndianess() != ExtEndian.CONTEXT) {
		    	res += "        bb.order(ctxbo);\r\n";
		    }
		}
		if(selpad > 0) {
			res += String.format("        bb.put(new byte[%d]);\r\n", selpad);
		}
        if(si.getEndpad()) {
		    int align = StructInfo.calculateEndAlign(si.getAlignment(), si.getFields());
		    if(align > 1) {
			    res += String.format("        npad = (%d - bb.position() %% %d) %% %d;\r\n", align, align, align);
		        res += "        if(npad > 0) bb.put(new byte[npad]);\r\n";
		    }
        }
		res += "    }\r\n";
		res += "}\r\n";
		log.fine("Generate:\r\n" + res);
		return res;
	}
	/**
	 * Generate FastStructWriter.
	 * @param <T> struct type
	 * @param t class of what to write
	 * @param bufsiz byte buffer
	 * @return FastStructWriter implementation
	 * @throws RecordException if impossible to convert between types in class and struct
	 */
	@SuppressWarnings("unchecked")
	public static <T> FastStructWriter<T> genWriter(Class<T> t, int bufsiz) throws RecordException {
        try {
        	String clznam = "dk.vajhoej.record.gen.StructWriterOf" + t.getSimpleName();
    		ClassPool cp = ClassPool.getDefault();
    		cp.importPackage("dk.vajhoej.record");
    		CtClass cc = cp.getOrNull(clznam);
    		if(cc == null) {
				cc = cp.makeClass(clznam);
				cc.addInterface(cp.get("dk.vajhoej.record.FastStructWriter"));
				cc.addField(new CtField(cp.get("java.nio.ByteBuffer"), "bb", cc));
				cc.addConstructor(CtNewConstructor.make("public " + t.getSimpleName() + "() { bb = java.nio.ByteBuffer.allocate(" + bufsiz + "); }", cc));
				cc.addMethod(CtNewMethod.make("public void workOn(java.nio.ByteBuffer bb) { this.bb = bb; }", cc));
				cc.addMethod(CtNewMethod.make( makeWriteMethod(t), cc));
				cc.addMethod(CtNewMethod.make("public byte[] getBytes() { int n = bb.position(); byte[] res = new byte[n]; bb.rewind(); bb.get(res); return res; }", cc));
				cc.addMethod(CtNewMethod.make("public java.nio.ByteBuffer getByteBuffer() { return bb; }", cc));
				return (FastStructWriter<T>) cc.toClass().newInstance();
    		} else {
    			return (FastStructWriter<T>)Class.forName(clznam).newInstance();
    		}
		} catch (NotFoundException e) {
			throw new RecordException("Error generating code for " + t.getName() + ": " + e.getMessage());
		} catch (InstantiationException e) {
			throw new RecordException("Error generating code for " + t.getName() + ": " + e.getMessage());
		} catch (IllegalAccessException e) {
			throw new RecordException("Error generating code for " + t.getName() + ": " + e.getMessage());
		} catch (CannotCompileException e) {
			throw new RecordException("Error generating code for " + t.getName() + ": " + e.getMessage());
		} catch (ClassNotFoundException e) {
			throw new RecordException("Error generating code for " + t.getName() + ": " + e.getMessage());
		}
	}
	private final static int DEFAULT_BUFSIZ = 10000;
	/**
	 * Generate FastStructWriter.
	 * @param <T> struct type
	 * @param t class of what to write
	 * @return FastStructWriter implementation
	 * @throws RecordException if impossible to convert between types in class and struct
	 */
	public static <T> FastStructWriter<T> genWriter(Class<T> t) throws RecordException {
		return genWriter(t, DEFAULT_BUFSIZ);
	}
}
