/*
 * Copyright 2021 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.record;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Map;

/**
 * Class CompositFastStructReader dispatches to a number of FastStructReaders's based on peek of actual sub class.
 */
public class CompositFastStructReader<T> implements FastStructReader<T> {
	private Map<Class<? extends T>, FastStructReader<? extends T>> map;
	private int seloffset;
	private Map<Integer, SubClassAndPad> selmap;
	private ByteBuffer bb;
	private ByteOrder bo;
	private FastStructReader<? extends T> fsr;
	/**
	 * Create instance of CompositFastStructReader.
	 * @param t super class
	 * @param map map from sub class'es to FastStructReader's
	 * @throws RecordException if problem with super class
	 */
	public CompositFastStructReader(Class<T> t, Map<Class<? extends T>, FastStructReader<? extends T>> map) throws RecordException {
		this.map = map;
		StructInfo si = StructInfoCache.analyze(t);
		int selfld = si.getSelectorField();
		seloffset = si.getOffset(si.getFields().get(selfld).getField().getName());
		selmap = si.getFields().get(selfld).getSelects();
		bo = si.getEndianess() == Endian.BIG ? ByteOrder.BIG_ENDIAN : ByteOrder.LITTLE_ENDIAN;
	}
	@Override
	public void init(byte[] ba) {
		bb = ByteBuffer.wrap(ba);
		bb.order(bo);
	}
	@Override
	public void init(ByteBuffer bb) {
		this.bb = bb;
		bb.order(bo);
	}
	@Override
	public T read(Class<T> t) throws RecordException {
		// peek at selector value to determine sub class
		int save = bb.position();
		bb.position(save + seloffset);
		int sel = bb.getInt();
		if(!selmap.containsKey(sel) ) {
			throw new RecordException("Unknown selector value: " + sel + " for " + t.getName());
		}
		@SuppressWarnings("unchecked")
		Class<T> clz = (Class<T>)selmap.get(sel).getSubClass();
		bb.position(save);
		// read instance of sub class
		fsr = map.get(clz);
		if(fsr == null) {
			throw new RecordException(clz.getName() + " is not mapped to a FastStructReader");
		}
		fsr.init(bb);
		@SuppressWarnings("unchecked")
		T res = ((FastStructReader<T>)fsr).read(clz);
		return res;
	}
	@Override
	public boolean more() {
		return fsr.more();
	}
	@Override
	public ByteBuffer getByteBuffer() {
		return fsr.getByteBuffer();
	}
}
